###############################################################################
#                                                                             #
#   Reproducible files to 'Random Forests for Ordinal Response Data' (2014)   #
#   by Janitza, Tutz and Boulesteix                                           #
#                                                                             #
#   Contact: S. Janitza <janitza@ibe.med.uni-muenchen.de>                     #
#                                                                             #
#   File for investigating if variables have any predictive accuracy          #
#                                                                             #
###############################################################################

# The results from this analyses are not shown in the paper. In this analysis
# we compare the performance of the fitted random forests to the performance 
# of a null model in order to investigate if the variables contained in the 
# original data have any predictive accuracy.

# please specify the path where the folder 'reproducible_files' is stored
path <- ""


# loading all datasets
# --------------------

# please load NHANES Data here:

load("")


# please load SUPPORT Study Data here:

load("support.sav")


# please load Mammography Experience Data here:

library(TH.data)
data(mammoexp)


# please load Very Low Birth Weight Data here:

load("vlbw.sav")


# please load Wine Quality Data here:

wine <- read.csv("winequality-white.csv", header = TRUE, sep = ";")




# data preprocessing
# ------------------

# preprocessing Wine Quality Data

wine <- wine[wine$quality < 9,] # only 5 observations with quality = 9 -> exclude



# preprocessing SUPPORT Study Data

# use only the following variables
support <- support[, names(support) %in% c("sfdm2", "age", "sex", "dzgroup", "dzclass", "num.co", "race", "charges", "scoma",                                           "avtisst", "meanbp", "wblc", "hrt", "resp", "temp", "crea", "sod")]

# exclude variables with more than 10% missing values
support <- support[, c(names(apply(support[, names(support) != "sfdm2"], 2, function(z) sum(is.na(z)) <= floor(0.1 * (nrow(support)-1)))), "sfdm2")]

# complete case analyses
support <- support[complete.cases(support),]



# preprocessing Very Low Birth Weight Data

# include only prenatal variables
vlbw <- vlbw[,names(vlbw) %in% c("apg1", "twn", "inout", "sex", "race", "delivery", "meth", "lol", "magsulf", "bwt", "toc")]

# convert variables to the correct data format
vlbw$twn     <- as.factor(vlbw$twn)
vlbw$magsulf <- as.factor(vlbw$magsulf)
vlbw$meth    <- as.factor(vlbw$meth)
vlbw$toc     <- as.factor(vlbw$toc)

# exclude observations with missing information
vlbw <- vlbw[complete.cases(vlbw),]

# exclude observations with too sparse levels
vlbw <- vlbw[vlbw$apg1 != 0,]                 # only 2 observations with apgar = 0
vlbw <- vlbw[vlbw$race != "oriental",]        # only 1 observation with race = oriental
vlbw <- vlbw[vlbw$race != "native American",] # only 4 observations with race = native American
vlbw$race <- factor(vlbw$race)




# RANDOM FOREST MODELS
# --------------------

# loading the R objects that contain the predictive accuracies of random forest models

setwd(paste(path, "reproducible_files/Real_Data_Applications/R_Objects/", sep = "")) # change working directory


# load all relevant R objects into the workspace

for(i in grep("pred_perf_", dir())) load(dir()[i])



# NULL MODEL
# ----------

# The null model is the model which does not contain any covariates but uses 
# only the information on the distribution of the response variable


pred_prob_mammoexp <- as.numeric(table(mammoexp$ME))/nrow(mammoexp)
pred_prob_NHANES   <- as.numeric(table(nhanes$HealthStatus))/nrow(nhanes)
pred_prob_SUPPORT  <- as.numeric(table(support$sfdm2))/nrow(support)
pred_prob_wine     <- as.numeric(table(wine$quality))/nrow(wine)
pred_prob_vlbw     <- as.numeric(table(vlbw$apg1))/nrow(vlbw)




# computed the performance (in terms of RPS) of the null model, which gives the same class probabilities to all observations 

compute_null_model_performance <- function(pred_prob, y){
  
  # compute the predicted cumulative distribution function 
  hat_F <- cumsum(pred_prob)
  
  ncat <- length(unique(y)) # number of response classes
  
  indicator_mat <- matrix(as.numeric(rep(y, each = ncat)), nrow = ncat, byrow = FALSE)
  T_F           <- apply(indicator_mat, 2, function(x) as.numeric(x <= 1:ncat))
    
  # compute and return RPS
  RPS <- sum((hat_F - T_F)^2)/length(y)
  
}

null_pred_perf_mammoexp <- compute_null_model_performance(pred_prob = pred_prob_mammoexp, y = mammoexp$ME)
null_pred_perf_NHANES   <- compute_null_model_performance(pred_prob = pred_prob_NHANES, y = nhanes$HealthStatus)
null_pred_perf_SUPPORT  <- compute_null_model_performance(pred_prob = pred_prob_SUPPORT, y = support$sfdm2)
null_pred_perf_wine     <- compute_null_model_performance(pred_prob = pred_prob_wine, y = wine$quality)
null_pred_perf_vlbw     <- compute_null_model_performance(pred_prob = pred_prob_vlbw, y = vlbw$apg1)



# SHOW RESULTS
# ------------

# Description of the plot:
#
#  - performance of models is measured by the RPS (the higher, the less good is the model)
#  - boxplots show the performance of the fitted random forest models (obtained through applying 10-fold cv)
#  - red line indicates the performance of a null model (fitted on the whole data)

graphics.off()
pdf(file = "../Figures/additional.pdf", width = 6, height = 8)

par(mar = c(13, 4, 4, 2))

boxplot(cbind(t(pred_perf_mammoexp[c(1, 3), ]), 
        t(pred_perf_NHANES[c(1, 3), ]), 
        t(pred_perf_SUPPORT[c(1, 3), ]), 
        t(pred_perf_wine[c(1, 3), ]), 
        t(pred_perf_vlbw[c(1, 3), ])),
        at = c(1, 2, 4, 5, 7, 8, 10, 11, 13, 14), ylab = "Models' prediction error measured by RPS", xaxt = "n", ylim = c(0.3, 1.55),
        main = "Comparison of prediction performance of \n random forest models with null model")

points(c(0.5, 2.5), c(null_pred_perf_mammoexp, null_pred_perf_mammoexp), col = "red", pch = 16, type = "l")
points(c(3.5, 5.5), c(null_pred_perf_NHANES, null_pred_perf_NHANES), col = "red", pch = 16, type = "l")
points(c(6.5, 8.5), c(null_pred_perf_SUPPORT, null_pred_perf_SUPPORT), col = "red", pch = 16, type = "l")
points(c(9.5, 11.5), c(null_pred_perf_wine, null_pred_perf_wine), col = "red", pch = 16, type = "l")
points(c(12.5, 14.5), c(null_pred_perf_vlbw, null_pred_perf_vlbw), col = "red", pch = 16, type = "l")

axis(side = 1, las = 2, at = c(1.5, 4.5, 7.5, 10.5, 13.5), labels = c("Mammography Experience Data", "NHANES Data", "SUPPORT Study", "Wine Quality Data", "Very Low Birth Weight Data"))

graphics.off()

