###############################################################################
#                                                                             #
#   Reproducible files to 'Random Forests for Ordinal Response Data' (2014)   #
#   by Janitza, Tutz and Boulesteix                                           #
#                                                                             #
#   Contact: S. Janitza <janitza@ibe.med.uni-muenchen.de>                     #
#                                                                             #
#   File containing relevant functions for real data studies                  #
#                                                                             #
###############################################################################


# ------------------------------------------------------------------------------
# Wrapper function for performing real data studies on prediction performance
# ------------------------------------------------------------------------------
#     input: * seed
#            * data
#            * yname: name of the response variable
#            * scores: vector of scores attached to each response levels, if not specified default scores are used
#            * mtry parameter for growing random forest, if not specified the default (sqrt(p), with p denoting the number of predictors) is used
#            * ncv: number for cross-validation (per default 10)
#     output: list with the following entries:
#               * accuracy of RF ordinal in terms of RPS
#               * accuracy of RF classification in terms of RPS
#               * accuracy of RF ordinal in terms of the error rate
#               * accuracy of RF classification in terms of the error rate
# ------------------------------------------------------------------------------


prediction_performance <- function(seed, data, yname, scores = NULL, mtry = NULL, ncv = 10){
  
  # set seed for reproducibility
  
  set.seed(seed)   
  
  
  # obtain number of categories
  
  ncat <- length(levels(data[, yname])) 
  
  
  # use default scores if no scores are specified
  
  if(!is.null(scores)){
    if(!length(scores) == ncat) stop("Score vector must have the same length as number of categories.") else scores <- eval(parse(text = paste("list(", yname, " = scores)")))
  }else{
    scores <- eval(parse(text = paste("list(", yname, " = 1:ncat)")))
  }
  
  
  # if mtry is not specified take the square root of the number of predictors 
  
  if(is.null(mtry)) mtry <- floor(sqrt(ncol(data) - 1))  
  
  
  # index for cross validation (while preserving class distribution)
  
  cvid <- rep(NA, length = nrow(data))
  for(i in levels(data[, yname])) cvid[data[, yname] == i] <- sample(rep(1:ncv, length = sum(data[, yname] == i)))

  
  # for each training set compute the accarcy of RF ordinal and RF classification and evaluate with the corresponding test set
  
  result <- sapply(1:ncv, function(cv){
    
    tryCatch({ # if there are technical problems in fitting the random forest, return NA
      
      # Fit random forests
      # ------------------
      
      RFs <- fit_RF(seed = seed, 
                    traindata = data[cvid != cv,],
                    yname = yname,
                    scores = scores,
                    mtry = mtry)
      
      
      # Compute prediction performance
      # ------------------------------
      
      PA_ordinal        <- get_PA(RF = RFs$RF_ordinal, testdata = data[cvid == cv,], yname = yname)
      PA_classification <- get_PA(RF = RFs$RF_classification, testdata = data[cvid == cv,], yname = yname)
      
      result <- c(PA_ordinal, PA_classification)
      
      
      names(result) <- c("RPS_RF_ordinal", "ER_RF_ordinal", "RPS_RF_classification", "ER_RF_classification")
      
      return(result)
      
    }, error = function(e){return(c(RPS_RF_ordinal = NA, ER_RF_ordinal = NA, RPS_RF_classification = NA, ER_RF_classification = NA))})
  }) 
  
  # return the mean value over the 10 repetitions (if problems in one or more iterations only use the results of the other iterations)
  
  return(rowMeans(result, na.rm = TRUE))
  
}



# ------------------------------------------------------------------------------
# Wrapper function for performing real data studies on variable importance
# ------------------------------------------------------------------------------
#     input: * seed
#            * data
#            * yname: name of the response variable
#            * scores: vector of scores attached to each response levels, if not specified default scores are used
#            * mtry parameter for growing random forest, if not specified the default (sqrt(p), with p denoting the number of predictors) is used
#     output: list with the following entries:
#               * accuracy of RF ordinal in terms of RPS
#               * accuracy of RF classification in terms of RPS
#               * accuracy of RF ordinal in terms of the error rate
#               * accuracy of RF classification in terms of the error rate
# ------------------------------------------------------------------------------


variable_importance <- function(seed, data, yname, scores = NULL, mtry = NULL){
  
  # set seed for reproducibility
  
  set.seed(seed)   
  
  
  # obtain number of categories
  
  ncat <- length(levels(data[, yname])) 
  
  
  # use default scores if no scores are specified
  
  if(!is.null(scores)){
    if(!length(scores) == ncat) stop("Score vector must have the same length as number of categories.") else scores <- eval(parse(text = paste("list(", yname, " = scores)")))
  }else{
    scores <- eval(parse(text = paste("list(", yname, " = 1:ncat)")))
  }
  
  # obtain number of predictor variables in the originally considered dataset
  
  p <- ncol(data) - 1 # -1 because of response variable
  
  
  # Add noise to data
  # -----------------
  
  # replicate predictor variables
  
  data <- data[, c(which(names(data) == yname), rep(which(names(data) != yname), 2))]
  
  
  # permute the block of replicated variables
  
  perm <- sample(1:nrow(data))
  data[, (p + 2) : (2*p + 1)] <- data[perm, (p + 2) : (2*p + 1)]
  
  
  
  # if mtry is not specified take the square root of the number of predictors 
  
  if(is.null(mtry)) mtry <- floor(sqrt(ncol(data) - 1))  
  
  
  # Fit random forests
  # ------------------
  
  RFs <- fit_RF(seed = seed, 
                traindata = data,
                yname = yname,
                scores = scores,
                mtry = mtry)
  
  
  # Compute and evaluate variable importance
  # ----------------------------------------
  
  AUC_ordinal        <- get_AUC(RF = RFs$RF_ordinal, scores = scores[[1]])
  AUC_classification <- get_AUC(RF = RFs$RF_classification, scores = scores[[1]])
  
    
  result <- c(AUC_ordinal, AUC_classification)
  
  names(result) <- c("AUC_ERVIM_RF_ordinal", "AUC_RPSVIM_RF_ordinal", "AUC_MSEVIM_RF_ordinal", "AUC_MAEVIM_RF_ordinal",
                     "AUC_ERVIM_RF_classification", "AUC_RPSVIM_RF_classification", "AUC_MSEVIM_RF_classification", "AUC_MAEVIM_RF_classification")

  return(result)
    
}


# ------------------------------------------------------------------------------
# Function for computing variable importance performance (in terms of AUC) 
# ------------------------------------------------------------------------------
#     input: * random forest object
#            * vector of scores attached to each response levels
#     output: vector with eight elements:
#               * performance of the error rate based VI for RF ordinal
#               * performance of the RPS-based VI for RF ordinal
#               * performance of the MSE-based VI for RF ordinal
#               * performance of the MAE-based VI for RF ordinal
#               * performance of the error rate based VI for RF classification
#               * performance of the RPS-based VI for RF classification
#               * performance of the MSE-based VI for RF classification
#               * performance of the MAE-based VI for RF classification
# ------------------------------------------------------------------------------


get_AUC <- function(RF, scores){
    
  VIM_ER  <- varimp(RF)                      # (standard) error rate based VI
  VIM_RPS <- varimpRPS(RF)                   # RPS-based VI
  VIM_MSE <- varimpMSE(RF, scores = scores)  # MSE-based VI (scores have to be specified)
  VIM_MAE <- varimpMAE(RF, scores = scores)  # MAE-based VI (scores have to be specified)
  
  effect_ind <- rep(c(1, 0), each = length(VIM_ER)/2)
  
  return(sapply(list(VIM_ER, VIM_RPS, VIM_MSE, VIM_MAE), function(z)
    unlist(performance(prediction(z, effect_ind), measure = "auc")@y.values)))
  
}




# ------------------------------------------------------------------------------
# Function for obtaining prediction accuracy
# ------------------------------------------------------------------------------
#     input: * seed
#            * training dataset (used to fit RF)
#            * test dataset (used to evaluate RF)
#            * vector of scores attached to each response levels
#            * mtry parameter for growing random forest
#     output: list with the following entries:
#               * accuracy of RF ordinal in terms of RPS
#               * accuracy of RF classification in terms of RPS
#               * accuracy of RF ordinal in terms of the error rate
#               * accuracy of RF classification in terms of the error rate
# ------------------------------------------------------------------------------


fit_RF <- function(seed, traindata, yname, scores, mtry){
  
  # set seed for reproducibility
  
  set.seed(seed)
  
  
  # specify model formula
  
  formula <- as.formula(paste(yname, "~", paste(names(traindata)[-which(names(traindata) == yname)], collapse = "+")))
  
    
  # specify response as ordinal variable 
  
  traindata[, yname] <- factor(traindata[, yname], ordered = TRUE)
  
  
  # Grow RF ordinal (random forest consisting of ordinal regression trees)
  # ----------------------------------------------------------------------
  
  RF_ordinal <- cforest(formula, 
                        data     = traindata, 
                        scores   = scores,
                        controls =
                          cforest_control(teststat     = "quad",
                                          testtype     = "Univ",
                                          mtry         = mtry,
                                          minbucket    = 0,
                                          minsplit     = 0,
                                          mincriterion = 0,
                                          replace      = FALSE,
                                          ntree        = 1000))
  
  
  # specify response as nominal variable 
  
  traindata[, yname] <- factor(traindata[, yname], ordered = FALSE)
  
  
  # Grow RF classification (random forest consisting of classification trees)
  # -------------------------------------------------------------------------
  
  RF_classification <- cforest(formula, 
                               data     = traindata, 
                               controls =
                                 cforest_control(teststat     = "quad",
                                                 testtype     = "Univ",
                                                 mtry         = mtry,
                                                 minbucket    = 0,
                                                 minsplit     = 0,
                                                 mincriterion = 0,
                                                 replace      = FALSE,
                                                 ntree        = 1000))
  
  
  return(list(RF_ordinal = RF_ordinal, RF_classification = RF_classification))
  
}



# ------------------------------------------------------------------------------
# Function for computing prediction accuracy (using ranked probability score and 
# error rate) 
# ------------------------------------------------------------------------------
#     input: * random forest object
#            * test dataset
#            * name of response variable
#     output: vector with two elements:
#             * first element is the ranked probability score
#             * second element is the error rate
# ------------------------------------------------------------------------------

get_PA <- function(RF, testdata, yname){
  
  ncat <- length(unique(RF@responses@variables[[1]]))
  
  # Ranked probability score
  # ------------------------
    
  # obtain predicted class probabilities and from that the predicted cumulative distribution function for each observation 
  
  pred_prob <- predict(RF, newdata = testdata, type = "prob")
  hat_F     <- sapply(pred_prob, function(x) cumsum(x))
  
  
  # obtain true cumulative distribution function for each observation
  
  indicator_mat <- matrix(as.numeric(rep(testdata[, yname], each = ncat)), nrow = ncat, byrow = FALSE)
  T_F           <- apply(indicator_mat, 2, function(x) as.numeric(x <= 1:ncat))
  
  
  # compute RPS
  
  RPS <- sum((hat_F - T_F)^2)/nrow(testdata)
  
  
  # Error rate
  # ----------
  
  pred_class <- as.numeric(sapply(pred_prob, which.max)) # obtain predicted classes 
  
  ER <- mean(as.numeric(pred_class) != as.numeric(testdata[, yname]))
  
  
  # return both measures
  
  return(c(RPS = RPS, ER = ER))
  
}

