###############################################################################
#                                                                             #
#   Reproducible files to 'Random Forests for Ordinal Response Data' (2014)   #
#   by Janitza, Tutz and Boulesteix                                           #
#                                                                             #
#   Contact: S. Janitza <janitza@ibe.med.uni-muenchen.de>                     #
#                                                                             #
#   File containing relevant functions for simulation studies                 #
#                                                                             #
###############################################################################


# ------------------------------------------------------------------------------
# Wrapper function for performing simulation studies
# ------------------------------------------------------------------------------
#
#     input: * seed
#            * ncat: number of response categories
#            * mixing: mixture proportion, per default 0.6 (in the paper termed as zeta)
#            * correlations: either TRUE for settings with correlated predictors or FALSE for settings with uncorrelated predictors only
#            * nobs: number of observations (in paper settings with n = 200 and n = 400)
#            * scores: vector of scores attached to each response levels, if not specified default scores are used
#            * get.tree.fractions.only: either FALSE (default), then the standard simulation is done, or TRUE then the fraction of trees is computed, in which an influential predictor was selected for a split (Figure 6)
#              The following arguments must not be changed:
#            * nnoise: number of noise predictors
#            * effect11: effect of predictor variables X_11, ..., X_15 on first mixture component
#            * effect12: effect of predictor variables X_6, ..., X_10 on first mixture component
#            * effect13: effect of predictor variables X_1, ..., X_5 on first mixture component
#            * effect2: absolute effect of predictor variables X_1, ..., X_4, X_6, ..., X_9, X_11, ..., X_14 on second mixture component
#            * corstrength: strength of correlation among signal predictors (only used if correlations = TRUE)
#            * ncor: number of correlated signal predictors (only used if correlations = TRUE)
#            * mtry: mtry parameter for growing random forest, if not specified the default (sqrt(p), with p denoting the number of predictors) is used
#
#     output: list with the following entries:
#               * accuracy of RF ordinal in terms of RPS
#               * accuracy of RF classification in terms of RPS
#               * accuracy of RF ordinal in terms of the error rate
#               * accuracy of RF classification in terms of the error rate
#               * performance of the error rate based VI for RF ordinal
#               * performance of the RPS-based VI for RF ordinal
#               * performance of the MSE-based VI for RF ordinal
#               * performance of the MAE-based VI for RF ordinal
#               * performance of the error rate based VI for RF classification
#               * performance of the RPS-based VI for RF classification
#               * performance of the MSE-based VI for RF classification
#               * performance of the MAE-based VI for RF classification
# ------------------------------------------------------------------------------

perform_simulation <- function(seed,
                               ncat         = 6,  
                               mixing       = 0.6,
                               correlations = TRUE,
                               nobs         = 200,
                               scores       = NULL,
                               nnoise       = 50, # must be more than 40
                               effect11     = 0.5, 
                               effect12     = 0.75, 
                               effect13     = 1, 
                               effect2      = 1,
                               corstrength  = 0.8,
                               ncor         = 6,
                               get.tree.fractions.only = FALSE,
                               mtry         = NULL){
  
  
  # use default scores if no scores are specified
  
  if(!is.null(scores)){
    if(!length(scores) == ncat) stop("Score vector must have the same length as number of categories.") else scores <- list(y = scores)
  }else{
    scores <- list(y = 1:ncat)
  }
  
  # if mtry is not specified take the square root of the number of predictors 
  
  if(is.null(mtry)) mtry <- floor(sqrt(nnoise + 15))
  
  
  # generate training dataset 
  
  traindata <- generate_data(seed = seed, n = nobs, ncat = ncat, mixing = mixing, correlations = correlations, nnoise = nnoise, ncor = ncor,
                             effect11 = effect11, effect12 = effect12, effect13 = effect13, effect2 = effect2, corstrength = corstrength)
  
  
  # check if one only wants to obtain the fraction of trees which select an important predictor for a first split
  
  if(get.tree.fractions.only){
    
    # obtain fraction of trees
    
    result <- view_first_split(seed = seed, traindata = traindata, mixing = mixing, scores = scores, mtry = mtry)
    
  }else{
    
    # generate large test dataset 
    
    testdata  <- generate_data(seed = seed * 10000, n = 10000, ncat = ncat, mixing = mixing, correlations = correlations, nnoise = nnoise, ncor = ncor,
                               effect11 = effect11, effect12 = effect12, effect13 = effect13, effect2 = effect2, corstrength = corstrength)
    
    
    # obtain results for generated training and test data using pre-specified scores
    
    result <- get_PA_VI(seed = seed, traindata = traindata, testdata = testdata, scores = scores, mtry = mtry, mixing = mixing)
  }
  
  return(result)
  
} 


# -----------------------------
# Function for data generation: 
# -----------------------------
#     input: * seed for reproducibility
#            * number of observations
#            * number of response categories
#            * mixture proportion
#            * indicator if correlated predictors or uncorrelated predictors are generated
#            * number of noise predictors
#            * number of correlated signal predictors
#            * effect of predictor variables X_1, ..., X_15 on both mixture components
#            * strength of correlation among signal predictors 
#     output: data frame
# -----------------------------

generate_data <- function(seed, n = nobs, ncat, mixing, correlations, nnoise, ncor, effect11, effect12, effect13, effect2, corstrength){
  
  # compute intercepts
  
  ncat_half <- ceiling(ncat/2)
  
  if(ncat %% 2 == 1){
    grid <- ncat_half*(ncat_half+1)/2-0.5
    thres <- 9/grid * c(0.5, 2:ncat_half)
    thresholds  <- c(-rev(cumsum(thres)), cumsum(thres))
  }else{
    grid <- ncat_half*(ncat_half+1)/2
    thres <- 9/grid * c(1:ncat_half)
    thresholds <- c(-rev(cumsum(thres)), 0, cumsum(thres))
  }
  
  thresholds[1]      <- -Inf # (per definition)
  thresholds[ncat+1] <- Inf  # (per definition)
  
  
  # coefficient vectors
  
  beta1 <- rep(c(effect13, effect12, effect11, 0), c(5, 5, 5, nnoise))
  beta2 <- c(rep(c(effect2, effect2, - effect2, - effect2, 0), length = 15), rep(0, nnoise))
  
  
  set.seed(seed)   # set seed for reproducibility
  
  # repeat data generation until there are enough observations in each response class
  tooless <- TRUE
  while(tooless){
    
    # Generating predictor matrix
    # ---------------------------
    
    if(correlations == TRUE){ # for settings with correlations between predictors
      
      # covariance matrix
      signal_sigma08 <- matrix(corstrength, nrow = ncor, ncol = ncor)  # 6 correlated signal predictors
      noise_sigma08  <- matrix(0.8, nrow = 10, ncol = 10)
      noise_sigma06  <- matrix(0.6, nrow = 10, ncol = 10)
      noise_sigma04  <- matrix(0.4, nrow = 10, ncol = 10)
      noise_sigma02  <- matrix(0.2, nrow = 10, ncol = 10)
      
      diag(signal_sigma08) <- diag(noise_sigma08) <- diag(noise_sigma06) <- diag(noise_sigma04) <- diag(noise_sigma02) <- 1
      
      # generate correlated predictors
      X_cor_signal   <- rmvnorm(n = n, sigma = signal_sigma08) 
      X_cor_noise08  <- rmvnorm(n = n, sigma = noise_sigma08) 
      X_cor_noise06  <- rmvnorm(n = n, sigma = noise_sigma06) 
      X_cor_noise04  <- rmvnorm(n = n, sigma = noise_sigma04)
      X_cor_noise02  <- rmvnorm(n = n, sigma = noise_sigma02)
      
      
      # generate uncorrelated predictors
      
      X_uncor_noise  <- rmvnorm(n = n, mean = rep(0, nnoise - 40))
      X_uncor_signal <- rmvnorm(n = n, mean = rep(0, 15 - ncor)) # 9 uncor. signal pred
      
      # combine predictors in a matrix
      
      X <- cbind(X_cor_signal[,1], X_uncor_signal[,1], X_cor_signal[,2], X_uncor_signal[,c(2, 3)],
                 X_cor_signal[,3], X_uncor_signal[,4], X_cor_signal[,4], X_uncor_signal[,c(5, 6)],
                 X_cor_signal[,5], X_uncor_signal[,7], X_cor_signal[,6], X_uncor_signal[,c(8, 9)],
                 X_cor_noise08, X_cor_noise06, X_cor_noise04, X_cor_noise02, X_uncor_noise)
      
      
    }else{ # for settings with uncorrelated predictors only
      X <- matrix(rnorm(n = n * (nnoise + 15)), nrow = n)
    }
    
    
    # Generating responses
    # --------------------
    
    # compute linear predictor
    
    eta1 <- apply(X %*% beta1, 1, function(z) z + thresholds[-(ncat + 1)])
    eta2 <- apply(X %*% beta2, 1, function(z) z + thresholds[-(ncat + 1)])
    
    
    # compute the probability p(Y <= r) from the mixture of two proportional odds models:
    # p(Y <= r) = zeta x p_1(Y <= r) + (1-zeta) x p_2(Y <= r)
    
    p_1 <- exp(eta1)/(1 + exp(eta1))                  # p_1(Y <= r)
    p_2 <- exp(eta2)/(1 + exp(eta2))                  # p_2(Y <= r)
    cumprob <- mixing * p_1 + (1 - mixing) * p_2      # p(Y <= r)
    
    
    # derive class probabilities from cumulative probabilities 
    
    cumprob <- rbind(cumprob, 1)       
    prob <- diff(cumprob)
    
    
    # perform multinomial experiments
    
    y_mat <- apply(prob, 2, function(z) rmultinom(n = 1, size = 1, prob = z))
    y     <- apply(y_mat, 2, which.max)
    
    # check if there are enough observations in each response class (at least 10 observations); if not, repeat data generation
    
    tooless <- min(table(y)) < 10
  }
  
  
  # return generated data
  
  return(data.frame(y = y, X = X))
  
}

# ------------------------------------------------------------------------------
# Function for obtaining prediction accuracy (PA) and variable importances (VI): 
# ------------------------------------------------------------------------------
#     input: * seed
#            * training dataset (used to fit RF and obtain variable importances)
#            * test dataset (used to evaluate RF)
#            * vector of scores attached to each response levels
#            * mtry parameter for growing random forest
#            * mixture proportion
#     output: list with the following entries:
#               * accuracy of RF ordinal in terms of RPS
#               * accuracy of RF classification in terms of RPS
#               * accuracy of RF ordinal in terms of the error rate
#               * accuracy of RF classification in terms of the error rate
#               * performance of the error rate based VI for RF ordinal
#               * performance of the RPS-based VI for RF ordinal
#               * performance of the MSE-based VI for RF ordinal
#               * performance of the MAE-based VI for RF ordinal
#               * performance of the error rate based VI for RF classification
#               * performance of the RPS-based VI for RF classification
#               * performance of the MSE-based VI for RF classification
#               * performance of the MAE-based VI for RF classification
# ------------------------------------------------------------------------------

get_PA_VI <- function(seed, traindata, testdata, scores, mtry, mixing){
  
  # set seed for reproducibility
  
  set.seed(seed)
  
  
  # specify response as ordinal variable 
  
  traindata$y <- factor(traindata$y, ordered = TRUE)
  
  
  # Grow RF ordinal (random forest consisting of ordinal regression trees)
  # ----------------------------------------------------------------------
  
  RF_ordinal <- cforest(y ~ ., 
                        data     = traindata, 
                        scores   = scores,
                        controls =
                          cforest_control(teststat     = "quad",
                                          testtype     = "Univ",
                                          mtry         = mtry,
                                          minbucket    = 0,
                                          minsplit     = 0,
                                          mincriterion = 0,
                                          replace      = FALSE,
                                          ntree        = 1000))
  
  
  # specify response as nominal variable 
  
  traindata$y <- factor(traindata$y, ordered = FALSE)
  
  
  # Grow RF classification (random forest consisting of classification trees)
  # -------------------------------------------------------------------------
  
  RF_classification <- cforest(y ~ ., 
                               data     = traindata, 
                               controls =
                                 cforest_control(teststat     = "quad",
                                                 testtype     = "Univ",
                                                 mtry         = mtry,
                                                 minbucket    = 0,
                                                 minsplit     = 0,
                                                 mincriterion = 0,
                                                 replace      = FALSE,
                                                 ntree        = 1000))
  
  
  

  
  # Compute prediction performance
  # ------------------------------
  
  PA_ordinal        <- get_PA(RF = RF_ordinal, testdata = testdata)
  PA_classification <- get_PA(RF = RF_classification, testdata = testdata)
  
  
  # Compute quality of all four variable importance measures
  # --------------------------------------------------------
  
  AUC_ordinal        <- get_AUC(RF = RF_ordinal, scores = scores$y, mixing = mixing)
  AUC_classification <- get_AUC(RF = RF_classification, scores = scores$y, mixing = mixing)
  
  
  # Return list with results on prediction accuracy and variable importance    
  # -----------------------------------------------------------------------
  
  result <- c(PA_ordinal, PA_classification, AUC_ordinal, AUC_classification)
  names(result) <- c("RPS_RF_ordinal", "ER_RF_ordinal", "RPS_RF_classification", "ER_RF_classification",
                     "AUC_ERVIM_RF_ordinal", "AUC_RPSVIM_RF_ordinal", "AUC_MSEVIM_RF_ordinal", "AUC_MAEVIM_RF_ordinal",
                     "AUC_ERVIM_RF_classification", "AUC_RPSVIM_RF_classification", "AUC_MSEVIM_RF_classification", "AUC_MAEVIM_RF_classification")
  
  return(result)
  
}




# ------------------------------------------------------------------------------
# Function for computing prediction accuracy (using ranked probability score and 
# error rate) 
# ------------------------------------------------------------------------------
#     input: * random forest object
#            * test dataset
#     output: vector with two elements:
#             * first element is the ranked probability score
#             * second element is the error rate
# ------------------------------------------------------------------------------

get_PA <- function(RF, testdata){
  
  
  # Ranked probability score
  # ------------------------
  
  ncat <- length(unique(testdata$y))  
  
  
  # obtain predicted class probabilities and from that the predicted cumulative distribution function for each observation 
  
  pred_prob <- predict(RF, newdata = testdata, type = "prob")
  hat_F     <- sapply(pred_prob, function(x) cumsum(x))
  
  
  # obtain true cumulative distribution function for each observation
  
  indicator_mat <- matrix(as.numeric(rep(testdata$y, each = ncat)), nrow = ncat, byrow = FALSE)
  T_F           <- apply(indicator_mat, 2, function(x) as.numeric(x <= 1:ncat))
  
  
  # compute RPS
  
  RPS <- sum((hat_F - T_F)^2)/nrow(testdata)
  
  
  # Error rate
  # ----------
  
  pred_class <- as.numeric(sapply(pred_prob, which.max)) # obtain predicted classes 
  
  ER <- mean(as.numeric(pred_class) != as.numeric(testdata$y))
  
  
  # return both measures
  
  return(c(RPS = RPS, ER = ER))
  
}


# ------------------------------------------------------------------------------
# Function for computing variable importance performance (in terms of AUC) 
# ------------------------------------------------------------------------------
#     input: * random forest object
#            * vector of scores attached to each response levels
#            * mixture proportion which indicates which predictors are influential
#     output: vector with eight elements:
#               * performance of the error rate based VI for RF ordinal
#               * performance of the RPS-based VI for RF ordinal
#               * performance of the MSE-based VI for RF ordinal
#               * performance of the MAE-based VI for RF ordinal
#               * performance of the error rate based VI for RF classification
#               * performance of the RPS-based VI for RF classification
#               * performance of the MSE-based VI for RF classification
#               * performance of the MAE-based VI for RF classification
# ------------------------------------------------------------------------------


get_AUC <- function(RF, scores, mixing){
    
  if(mixing == 0) effect_ind <- c(rep(c(1, 1, 1, 1, 0), length = 15), rep(0, 50)) else effect_ind <- rep(c(1, 0), c(15, 50))
  
  VIM_ER  <- varimp(RF)                      # (standard) error rate based VI
  VIM_RPS <- varimpRPS(RF)                   # RPS-based VI
  VIM_MSE <- varimpMSE(RF, scores = scores)  # MSE-based VI (scores have to be specified)
  VIM_MAE <- varimpMAE(RF, scores = scores)  # MAE-based VI (scores have to be specified)
  
  return(sapply(list(VIM_ER, VIM_RPS, VIM_MSE, VIM_MAE), function(z)
    unlist(performance(prediction(z, effect_ind), measure = "auc")@y.values)))
  
}




# ------------------------------------------------------------------------------
# Function for obtaining the fraction of trees of RF ordinal or RF classificati-
# on, in which a signal predictor was selected for the first split: 
# ------------------------------------------------------------------------------
#     input: * seed
#            * training dataset
#            * mixture proportion which indicates which predictors are influential
#            * vector of scores attached to each response levels
#            * mtry parameter for growing random forest
#     output: vector with two elements:
#             * first element is tree fraction of RF ordinal 
#             * second element is tree fraction of RF classification
# ------------------------------------------------------------------------------

view_first_split <- function(seed, traindata, mixing, scores, mtry){
  
  # set seed for reproducibility
  
  set.seed(seed)
  
  
  # check which predictors have an effect (depends on the considered scenario)
  
  if(mixing == 0) effect.pred <- c(1:4, 6:9, 11:14) else effect.pred <- 1:15
  
  
  # sample mtry predictors and 0.632 of the observations (1000 times), return indices for both
  
  index <- lapply(1:1000, function(x) list(sample(1:(ncol(traindata)-1), size = mtry), sample(1:nrow(traindata), size = floor(0.632 * nrow(traindata)))))  
  
  
  # Function for fitting an ordinal regression tree and a classification tree; for both trees the predictor is returned that was selected for the first split
  # this strategy ensures that the same mtry predictors and observations are selected for fitting the classification and ordinal regression trees
  
  grow_trees <- function(pred.index, obs.index){
  
    # specify response as ordinal variable 
    
    traindata$y <- factor(traindata$y, ordered = TRUE)
    
    
    # fit ordinal regression tree (only stumps are required)
    
    tree_ordinal <- ctree(y ~ ., data = traindata[obs.index, c(1, (pred.index + 1))], # use only mtry randomly drawn predictors and a subsample of the data
                          controls = ctree_control(teststat       = "quad",
                                                   testtype       = "Univ",
                                                   minbucket      = 0,
                                                   minsplit       = 0,
                                                   mincriterion   = 0,
                                                   savesplitstats = FALSE,
                                                   stump          = TRUE))
    
    
    # specify response as nominal variable 
    
    traindata$y <- factor(traindata$y, ordered = FALSE)
    
    
    # fit classification tree (only stumps are required)
    
    tree_classification <- ctree(y ~ ., data = traindata[obs.index, c(1, (pred.index + 1))],  # use only mtry randomly drawn predictors and a subsample of the data
                                 controls = ctree_control(teststat       = "quad",
                                                          testtype       = "Univ",
                                                          minbucket      = 0,
                                                          minsplit       = 0,
                                                          mincriterion   = 0,
                                                          savesplitstats = FALSE,
                                                          stump          = TRUE))
    
    
    # extract the predictor selected for the first split of the tree
    
    return(c(tree_ordinal@tree@.Data[[5]]$variableName, tree_classification@tree@.Data[[5]]$variableName))
  }
  
  # extract predictors selected for the first split for 1000 ordinal regression trees (RF ordinal) and 1000 classification trees (RF classification)
  
  sel.pred <- sapply(index, function(z) grow_trees(pred.index = z[[1]], obs.index = z[[2]]))
  
  
  # assess tree fraction where signal predictors were selected for the first split 
  
  obtain_tree_frac <- function(pred) mean(pred %in% paste("X.", effect.pred, sep = ""))
  fraction <- apply(sel.pred, 1, obtain_tree_frac)
  
  return(fraction)  # return fractions
}
